RequirePackage("require_file");
AddToLoadPath("~/research/gap-lrslib/src");
AddToLoadPath("../wreath");

ReadFile("dfs.gap");

Rot:=[
      [1, 0, 0],
      [0, 0, -1],
      [0, 1, 0]
      ];


Neg:=[ [1, 0, 0,],
      [ 0, -1, 0],
      [ 0, 0, 1] ];

G1:=Group([Rot,Neg]);

Octa:=SortedList(Orbit(G1,[1,2,1]));

G2:=Group(Rot);
Square:=SortedList(Orbit(G2,[1,1,1]));


zeroRow:=function(d)
    return Concatenation([1], List([1..d],i->0));
end;

rowMat:=function(v,n,k)
    return Concatenation(
                   List([1..k-1], i->zeroRow(Size(v)-1)),
                   [v],
                   List([k+1..n], i->zeroRow(Size(v)-1)));
end;


homConcat:=function(L)
    return Concatenation([1],
                   Concatenation(List(L,r->r{[2..Size(r)]})));
end;

wreath:=function(P,Q)
    local V,i,k;
    V:=[];
    
    for i in [1..Size(P)] do
        for k in [1..Size(Q)] do
            Add(V,homConcat([homConcat(rowMat(P[i],Size(Q),k)),Q[k]]));
        od;
    od;
    return V;
end;

WP:=wreath(Square,Octa);

liftTrans:=function(M,n,k)
    local I, F,d;
    d:=Size(M);
    I:=IdentityMat(n);
    F:=function(i,j)
        # skip first column, row
        if (i <= k or j<= k  or i>= k+d or j>=k+d) then 
            return I[i][j];
        else
            return M[i-k+1][j-k+1];
        fi;
    end;
    return List([1..n],i->List([1..n],j->F(i,j)));
end;

PG4:=Group((1,2,3,4));
PG8:=Group((1,3,5,7)(2,4,6,8),(1,8)(2,7)(3,6)(4,5));


moveCols:=function(pi) 
    return PermList(Concatenation([1],Concatenation(List(ListPerm(pi),j->[2*j,2*j+1]))));
end;

# note this is really the generators of PG8, but relabelled because the points are not generated in the 
# orbit in a nice rotational order.
pi:= moveCols((1,4,8,5)(2,6,7,3));
mu:=moveCols((4,6)(2,8)(1,7)(3,5));

Rot2:= PermutationMat(pi,Size(WP[1]))*liftTrans(Rot,19,17);
Neg2:= PermutationMat(mu,Size(WP[1]))*liftTrans(Neg,19,17);

Gens:=List(Concatenation(List([1,3..15],k->liftTrans(Rot,19,k)),[Rot2,Neg2]));

WG:=Group(Gens);



globalResults:=rec();
doDFS:=function()
    local results;
    ClearProfile();
    ProfileFunctions(dfs,IsNewCobasis,StabCanon,FindSymmetry,knownVertex,
            MatchingInvariants,CanonicalizeCobasis,CobasisInvariants,PushNewEdges);
    globalResults:= dfs(WP,fail,
                        [ rec(cobasisStabilizers:=false, 
                              lazyCanon:=false,
                              gramVec:=true,
                              gramGraph:=true,
                              gramMotion:=true,
                              gramMotionViaBrute:=true,
                              trustCanon:=false,
                              lexOnly:=false,
                              infoLevel:=1,
                              siftThreshhold:=10000,
                              greedyTrials:=200,
                              )]);;
    DisplayProfile();
end;
