#!/usr/bin/env perl

use strict;
use File::Temp qw/ tempdir /;
use Algorithm::Diff qw/diff/;
use Data::Dumper qw/Dumper/;
use File::Slurp;

my $binfile="../lrsserv/lrsserv";

my $tempdir=tempdir(CLEANUP => 0);

my $testcount=0;
my $failcount=0;
while (<*.out>){
  my $in=$_;
  my $out=$_;
  
  $testcount++;
  print STDERR "# test $testcount: $in\n";
  $in =~ s/\.out/\.in/;
  my $tmpout=$tempdir."/".$out;
  if (system($binfile,$in,$tmpout)){
    $failcount++;
    print STDERR "CRASHED\n";
  } else {
    undef $/;
    my @canonout=read_file($out);
    my @runout=read_file($tmpout);
    my $diff=diff (\@canonout,\@runout);
    if (scalar(@{$diff})>0) {
      $failcount++;
      print STDERR "FAILED\n";
      print Dumper($diff);
    } else {
      print STDERR "OK\n";
    }
  }
}

print STDERR "$testcount Tests\n";
print STDERR "$failcount Failures\n";
