# -*- Gap -*- 

BindGlobal("NAUTY_LONG_TIMEOUT",3600);
BindGlobal("NAUTY_SHORT_TIMEOUT",1);
BindGlobal("NautyDebug",0);

InstallGlobalFunction(Nauty,
        function()

    local output,process;
    process:= InputOutputLocalProcess(DirectoryCurrent(),
                      Filename(DirectoriesPackagePrograms("nauty"),"nautyserv"),[]);
    if (process=fail) then
	Info(InfoWarning,1, "Failed to start nautyserv");
        return fail;
    else	
        if (Expect(process,"NAUTYSERV HELLO")) then
            return process;
        else
            return fail;
        fi;	
    fi;
end);	

InstallGlobalFunction(NautyLoadGraph,
        function(p,G)
    local i,L,part;
    
    Info(InfoNauty,4,"entering NautyLoadGraph ",Runtime());
    if IsBound(G.ident) then
        SendLine(p,G.ident,"ident");
    fi;
    SendLine(p,G.numVertices,"graph");
    Insist(p,"OK");
    
    L:=[];
    
    for i in [1..Size(G.adjList)] do
        L:=Concatenation(L,["@"],G.adjList[i],[i],["adjrow\n"]);
    od;
    
    SendList(p,L);
    

    Info(InfoNauty,4,"edges sent ",Runtime());    
    part:=ShallowCopy(G.partition);
# don't sort; we need to preserve the implicit labels.
#    Sort(part,function(a,b) return (Minimum(a) < Minimum(b)); end);
    for i in [1..Size(part)] do
        SendItems(p,"@");
        SendList(p,part[i]);
        SendLine(p,"colourclass");
        Insist(p,"OK");
    od;
    Info(InfoNauty,4,"leaving NautyLoadGraph ",Runtime());    
    return true;
end);

InstallMethod(NautyGraph,[IsPosInt,IsList,IsList,IsInt],
        function(n,adj,colours,ident)
    local indices,partition,uniqueColours,i;
    
    uniqueColours:=AsSet(colours);
    
    partition:=List(uniqueColours,p->[]);
    
    for i in [1..Size(colours)] do
        Add(partition[Position(uniqueColours,colours[i])],i);
    od;

    return rec(numVertices:=n,
               adjList:=adj,
               colours:=colours,
               partition:=partition,
               ident:=ident);
end);


InstallMethod(NautyRunNauty,[IsInputOutputStream],
        function(process)
#    SendLine(process,"dumpgraph");
#    SendLine(process,"dumppartition");
    SendLine(process,"nauty");
    Insist(process,"DONE",NAUTY_LONG_TIMEOUT);
end);

InstallMethod(NautyFindAutomorphisms,[IsInputOutputStream],
        function(process)
    local line,gens,list;
    
    SendLine(process,"printautom 1 option");
    if NautyDebug>0 then
        SendLine(process,"dumpgraph");
        SendLine(process,"dumppartition");
    fi;
    SendLine(process,"nauty");
    gens:=[];
    repeat
        line:=ReadLineWithTimeout(process,NAUTY_LONG_TIMEOUT);
        Info(InfoNauty,3,"Nauty returned ",line);
        if line=fail then 
            return fail;
        fi;
        line:=Chomp(line);
        if (line <> "DONE") then
            list:=StringToRatVec(line);
            Add(gens,PermList(list));
        fi;
    until line="DONE";
    return gens;
end);

InstallMethod(NautyGetCanon,[IsInputOutputStream],
        function(process)
    local line;
    SendLine(process,"getcanon");
    line:=ReadLineWithTimeout(process,NAUTY_SHORT_TIMEOUT);
    if line=fail then
        return fail;
    fi;
    return PermList(StringToRatVec(line));
end);


InstallMethod(NautyRUsage,[IsInputOutputStream],
        function(process)
    local line,gens,list,key,rusage;
    
    SendLine(process,"rusage");
    rusage:=rec();
    repeat
        Info(InfoNauty,1,"Going to read");        
        line:=ReadLineWithTimeout(process,NAUTY_LONG_TIMEOUT);

        Info(InfoNauty,1,"Nauty returned ",line);
        if line=fail then 
            return fail;
        fi;
 	key:=Chomp(line);
        if (key <> "DONE") then
            Info(InfoNauty,1,"Got key ", key);
            line:=ReadLineWithTimeout(process,NAUTY_LONG_TIMEOUT);	
            if line=fail then 
                return fail;
            fi;

            list:=StringToRatVec(line);
            # convert to milliseconds; some loss of precision
            if (key="user_time" or key="system_time") then
                rusage.(key):=RoundCyc(list[1]*1000+list[2]/1000);
            elif (key="memory") then
                 rusage.(key):=list[1];
            else
                rusage.(key):=list;
            fi;
        fi;
    until key="DONE";
    return rusage;
end);
