/* Based on Brendan McKays example program nautex3.c
*/

#include <sys/resource.h>
#include <errno.h>
#include "nauty.h"    /* which includes <stdio.h> */
#include "naugroup.h"
#include "nautyserv.h"
#include "thingstack.h"
static void putline(char *);
static void put_int_vec(int *, int);
static void put_int_pair(int, int);
static void retcode(int);
static void putline(char *);
static void putint(int);
static void endline();

FILE *debugfp;
DYNALLSTAT(graph,g,g_sz);
DYNALLSTAT(graph,canong,canong_sz);
DYNALLSTAT(int,lab,lab_sz);
DYNALLSTAT(int,ptn,ptn_sz);
DYNALLSTAT(int,orbits,orbits_sz);

static DEFAULTOPTIONS_GRAPH(options);
statsblk(stats);
static int n,m,lastlabel;

setword workspace[100];


void
automproc(int count, permutation *p, int *orbits, int numorbits, 
	  int stabvertex, int n)
{
    int i;


    putint(n);
    for (i = 0; i < n; ++i) 
      putint(p[i]+1);

    endline();
}

/**************************************************************************/


void newgraph(){
  int v;

  dlog("K>=graph");
  n=pop_int();
  m = (n + WORDSIZE - 1) / WORDSIZE;
  nauty_check(WORDSIZE,m,n,NAUTYVERSIONID);
  
  DYNALLOC2(graph,g,g_sz,m,n,"malloc");
  DYNALLOC2(graph,canong,canong_sz,m,n,"malloc");
  DYNALLOC1(int,lab,lab_sz,n,"malloc");
  DYNALLOC1(int,ptn,ptn_sz,n,"malloc");
  DYNALLOC1(int,orbits,orbits_sz,n,"malloc");
  lastlabel=0;

  for (v=0; v<n ; v++){
    set *gv=GRAPHROW(g,v,m);
    EMPTYSET(gv,m);
  }
    
  putline("OK");
}

/* WARNING: if nauty is given a partition not ordered by the smallest
   element, wierd things happen. This is not checked here, but is
   checked in the gap glue code.
*/
void colourclass(){
  int u;
  
  dlog("K=>colourclass");
  options.defaultptn=0;

  while (!top_is_mark()){
    u=pop_int()-1;
    
    lab[lastlabel]=u;
    ptn[lastlabel]=1;
    lastlabel++;
  };
  ptn[lastlabel-1]=0;
  pop();
  

  putline("OK");
}

void adjrow(){
  int v,j,u;
  set *gv;

  dlog("K=>adjrow");
  v=pop_int()-1;
  
  gv = GRAPHROW(g,v,m);

  EMPTYSET(gv,m);

  while (!top_is_mark()){
    u=pop_int()-1;
    ADDELEMENT(gv,u);
  };
  pop();
}
  
extern void donauty(){
  dlog("K=>nauty");
  nauty(g,lab,ptn,NULL,orbits,&options,&stats,
	workspace,100,m,n,canong);
  putline("DONE");
}
extern void shutdown(){
  exit(0);
}

static void put_int_vec(int *v, int n){
  int j;
  putint(n);
  for (j=0; j<n; j++){
    putint(v[j]);
  }
  endline();
}

static void put_int_pair(int a, int b){
  int j;
  putint(2);
  putint(a);
  putint(b);
  endline();
}

static void endline(){
  dlog("<=\\n");
  fputs("\n",stdout);
  fflush(stdout);
}

static void putint(int i){
  dlog("<=%d",i);
  printf("%d ",i);
}

static void putline(char *s){
  dlog("<=%s",s);
  puts(s);
  fflush(stdout);
}

void dumpgraph(){
  fprintf(stderr,"n=%d $=%d g\n",n,labelorg);
  putgraph(stderr,g,options.linelength,m,n);
}


void dumppartition(){
  int i;
  fprintf(stderr,"f=");
  putptn(stderr,lab,ptn,0,options.linelength,n);
}



void getcanon(){
  int j;
  putint(n);
  for (j=0; j<n; j++){
    putint(lab[j]+1);
  }
  endline();
}

void option(){
  int tok,val;

  val=pop_int();
  tok=pop_int();
  
  switch(tok) {
  case TOK_PRINTAUTOM:
    if (val) {
        options.userautomproc=automproc;
	dlog("O== printautom");
    }
    break;
  default:
    err_fatal("option ", "unknown token %d", tok);
  }
}

void rusage(){
  struct rusage ru_s;
  if (getrusage(RUSAGE_SELF,&ru_s)){
    err_fatal("getrusage: %s",strerror(errno));
  }
  putline("user_time");
  put_int_pair(ru_s.ru_utime.tv_sec,ru_s.ru_utime.tv_usec);
  putline("system_time");
  put_int_pair(ru_s.ru_stime.tv_sec,ru_s.ru_stime.tv_usec);    
  putline("memory");
  putint(1);
  putint(ru_s.ru_maxrss);
  endline();
  putline("DONE");
}
int
main(int argc, char *argv[])
{
  extern FILE *yyin;
  if (argc>1){
    yyin=fopen(argv[1],"r");
  }

#ifdef DEBUGLOG
  {
    char buf[BUFSIZ];
    sprintf(buf,"/tmp/nautyserv%d.log",getpid());
    debugfp=freopen(buf,"w",stderr);
    setvbuf(debugfp,NULL,_IOLBF,BUFSIZ);
  }
#endif
  

  options.getcanon=1;


  setvbuf(stdout,NULL,_IOLBF,BUFSIZ);
  putline("NAUTYSERV HELLO");
  yylex();
  return 0;
}
