RequirePackage("lrslib");
RequirePackage("nauty");
RequirePackage("require_file");



InfoDFS:=NewInfoClass("DFS");
SetInfoLevel(InfoDFS,1);

RequireFile("sort-dict.gap");
RequireFile("profile.gap");
RequireFile("canon.gap");
RequireFile("perm-mat-group.gap");
RequireFile("gram.gap");
RequireFile("ratio.gap");

#RequirePackage("grape");



#option:=function(options,key)
#    if key in RecNames(options) then
#        return options.(key);
#    else
#        return false;
#    fi;
#end;

ErrCheck:=function(name,val) 
    if (not val) then 
        Error("Function ",name, " failed");
    fi;
end;

######################################################################
# Maintain two stacks of pivots, one to go down, and another to go up.

PushPivot:=function(S,C,l,e)
    Add(S,rec(cobasis:=C,leave:=l,enter:=e));
end;

PopPivot:=function(S)
    local last;
    last:=S[Size(S)];
    Unbind(S[Size(S)]);
    return last;
end;

DequeuePivot:=function(S)
    local last;
    last:=S[1];
    Remove(S,1);
    return last;
end;

######################################################################
# do a pivot on the index set only

CombPivot:=function(C,l,e)
    local output;
    output:=SortedList(C);
    RemoveSet(output,l);
    AddSet(output,e);
    return output;
end;

######################################################################

######################################################################
# Gram Matrices

                   

OnGraphMat:=function(A,g)
    local rows;
    rows:=List(A,r->Permuted(r,g));
    
    return Permuted(rows,g);
end;



histogram:=function(v)
    local sorted,j,last,count,hist;
    sorted:=AsSortedList(v);
    last:=sorted[1];
    count:=0;
    hist:=[];
    for j in sorted do
        if j <> last then
            Add(hist,[last,count]);
            count:=0;
        fi;
        last:=j;
        count:=count+1;
    od;        
    Add(hist,[last,count]);
    return hist;
end;


colourMap:=function(d,edges,M)
    local colours,e,c,D;
    
    D:=newDictionary();
    colours:=[];

    for e in edges do
        c:=M[e[1]][e[2]];
        AddSet(colours,c);
        addDictionary(D,c,Position(edges,e)+d);
    od;
    return Concatenation([[1..d]],
                       List(colours,c->AsSet(lookupDictionary(D,c))));
end;

        
#colourOf:=function(G,v,M)
#    local edge;
#    edge:=VertexName(G,v);
#    if Size(edge)=2 then 
#        return M[edge[1]][edge[2]];
#    else
#        return 0;
#    fi;
#end;

######################################################################

LinearMapPerm:=function(pi,A)
    local iCols,dCols,Aprime,Atranspose,dim,B,T;
    
    dim:=Size(A[1]);
    
    iCols:=LinearIndependentColumns(A);
    dCols:=Difference([1..Size(A[1])],iCols);
    
    # Augment A to be full column rank.
    Aprime:=Concatenation(A,IdentityMat(dim){dCols});
    
    Atranspose:=TransposedMat(Aprime);
    
    B:=Aprime{Permuted([1..Size(Aprime)],pi)};
    
    # do the left inverse trick.  
#    Print(1/(Atranspose*Aprime));
    return (1/(Atranspose*Aprime))*(Atranspose*B);
end;

PermLinearMap:=function(T,M)
    local newM;
    newM:=M*T;
#    Info(InfoDFS,2,newM);
    return PermList(List(M,v->Position(newM,v)));
end;


######################################################################

# Note that we need to find how the transformation acts on _all_ of the facets,
# not just those that happen to be in the cobasis.
recoverPerm:=function(cob,pi,M)
    local A,iCols,B1,B2,Aprime,T,L;
    
    iCols:=LinearIndependentColumns(M);
    
    
    A:=List(M,r->r{iCols});
    
    
    B1:=A{cob};
    B2:=Permuted(B1,pi);
    if (Size(cob) < Size(iCols)) then
        # this is not that well thought out...
        L:=Concatenation([1],List(cob, i->0));
        Print(L);
        Add(B1,L);
        Add(B2,L);
    fi;
    
    Print(B1);
    Print(B2);
    T:=(1/B1)*B2
;
    
    Aprime:=A*T;
    
    return List(Aprime,r->r in A);
    
end;




wrapCanonGramViaNauty:=function(G,cob,globals)
    local ident;
    
    Info(InfoDFS,4,"raw graph:",G);
    
    ident:=globals.nextIdent;
    globals.nextIdent:=globals.nextIdent+1;
    
    return canonGramViaNauty(G,cob,ident,globals.Nauty);
    
    
end;


AugmentGraphMat:=function(M)
    return Concatenation(List(M,r->Concatenation(r,[0])),
                   [Concatenation([1..Size(M)],[0])]);
end;

DeaugmentGraphMat:=function(A) 
    return List(A{[1..Size(A)-1]},r->r{[1..Size(A)-1]});
end;

AugMatCompare:=function(a,b)
    return DeaugmentGraphMat(a) < DeaugmentGraphMat(b);
end;

IndexifyGroup:=function(G,cob)
    local rgen, pi;
    
    rgen:=[];
    for pi in GeneratorsOfGroup(G) do
        Add(rgen,PermListList(cob,OnTuples(cob,pi)));
    od;
    return Group(rgen);
end;
          
gramGraph:=function(rep,globals,options)
    local M,d,gg,G,cob,indexList,A,S;
    
    cob:=rep.cobasis;
    
    d:=Size(cob);
    M:=fastGramMat(cob,globals.innerProductMat);
    if options.gramViaStabilizer then
        A:=AugmentGraphMat(M);
        S:=IndexifyGroup(rep.stabilizer,cob);
        A:=ExhaustiveCanon(S,A,OnGraphMat,AugMatCompare);
        M:=DeaugmentGraphMat(A);
        indexList:=A[Size(A)]{[1..Size(M)]};
    fi;
    
    
    G:=buildGramGraph(rep.cobasis,M);
    if options.gramViaStabilizer then
        gg:=NautyGraph(G.numVerts,G.adjList,G.colours);
        gg.labels:=cob{indexList};
        gg.automorphismGroup:=rep.stabilizer;
    else
        gg:=wrapCanonGramViaNauty(G,rep.cobasis,globals);
    fi;
    
    Info(InfoDFS,4,"canonicalized graph:",gg);
    return gg;
end;



gramKey:=function(dim,graph)
    local edges,i;
    edges:=graph.adjList{[dim+1..Size(graph.adjList)]};
    edges:=List(edges,ShallowCopy);
    for i in [1..Size(edges)] do
        Add(edges[i],graph.colours[dim+i]);
    od;
    Sort(edges);
    return Concatenation(edges);
end;

######################################################################
# Isometry testing


WidthList:=function(L)
    return Maximum(L);
end;


SetCompare:=function(a,b)
    if Maximum(a) < Maximum(b) then
        return true;
    elif Maximum(b) < Maximum(a) then
        return false;
    else
        return a<b;
    fi;
end;



PseudoCanon:=function(element,globals,options)
    local O,r,stab,heuristic,canon;
    
    heuristic:=GreedyDescent(globals.PermGrp,element,
                       OnSets,SetCompare,options.greedyTrials);
    Info(InfoDFS,3,"heuristic canon = ",heuristic,"\n");
    canon:=StabOrbitMinimum(globals.PermGrp,element,SetCompare);
    Info(InfoDFS,2,"Canon rep ",canon);
    return canon;
end;


BrutalCanon:=function(element,globals)
    return StabCanon(globals.PermGrp,element,globals.dim+1,globals.rows,
                   globals.pseudoCanons);
end;

LazyCanon:=function(element,globals,options)
    local heuristic;
    heuristic:=GreedyDescent(globals.PermGrp,element,OnSets,
                       SetCompare,options.greedyTrials);
                       
                       
    element:=StabCanon(globals.PermGrp,heuristic,
                     globals.dim+1,globals.groundMax,
                     globals.pseudoCanons);
    return element;
end;




realSiftCanons:=function(group, canons, threshhold)
    local newCanons,canon;
    
    newCanons:=[];
    
    Info(InfoDFS,1,"Sifting  ",Size(canons),
         " canonical representatives with moved= ", threshhold);
    

    for canon in canons do
        AddSet(newCanons,
               StabCanon(group,canon,threshhold,
                       threshhold,newCanons));
        Info(InfoDFS,4,"Found ",Size(newCanons));
    od;
    Info(InfoDFS,1,"result  ",Size(newCanons), " representatives");
    
    return newCanons;

end;

# note, this should really clean out the basis records if trustCanon is set

SiftCanons:=function(globals)
    
    if globals.groundMax< globals.rows then
        globals.groundMax:=globals.groundMax+1;
    else
        # nothing to do
        return;
    fi;
    
    globals.pseudoCanons:=realSiftCanons(globals.PermGrp,
                                  globals.pseudoCanons, 
                                  globals.groundMax);
    
    while (globals.siftThreshhold <= 2*Size(globals.pseudoCanons)) do
        globals.siftThreshhold:= globals.siftThreshhold*2;
    od;
end;            

# assumes both matrices have full row rank
# and that some set of linear independent columns is common
findLinearMap:=function(L1,L2,M)
    local A,B1,B2,iCols,pi;
    
    Info(InfoDFS, 4, "checking for lin map from ",L1," to ", L2);
# iCols:=LinearIndependentColumns(M{L1});
    
    # FIXME this is a temporary hack, and assumes first column is all ones.
    iCols:=[2..Size(M[1])];
    A:=List(M,r->r{iCols});
    
    
    B1:=A{L1};
    B2:=A{L2};
    
    # B1 = B2*P
    P:= (1/B2)*B1;
    
    pi:= PermListList(A*P,A);
    Info(InfoDFS,4,"Linear map =",pi);
    return pi;
end;

isIsometry:=function(pi,globals)
    local i,j,M,IPM;
    
    M:=globals.Mat;
    IPM:=globals.innerProductMat;
    
    if pi=fail then 
        return false;
    elif lookupCache(globals.isometryCache,pi) then
        Info(InfoDFS,4,"found isometry in cache ");
        return true;
    fi;
        
    for i in [1..Size(M)] do
        for j in [i..Size(M)] do
            if IPM[i][j] <> IPM[OnPoints(i,pi)][OnPoints(j,pi)] then
                Info(InfoDFS,4,"failing isometry test ");                
                return false;
            fi;
        od;
    od;
    addCache(globals.isometryCache,pi);
    Info(InfoDFS,4,"found new isometry");
    return true;
end;

addIsometry:=function(pi,globals)
    globals.PermGrp:=ClosureGroup(globals.PermGrp,pi);
end;

gramMotion:=function(new,old,globals,options)
    local I,A,motion,pi,M;
    
    M:=globals.Mat;
    
    for pi in AsList(new.gramGraph.automorphismGroup) do
        I:=OnTuples(new.gramGraph.labels,pi);
        Info(InfoDFS,4,"gramMotion trying ",I);
        motion:=findLinearMap(I,old.gramGraph.labels,M);
        if (options.gramMotionViaGroup) then
            if motion in globals.PermGrp then
                return motion;
            fi;
        elif (options.gramMotionViaBrute) then
            if isIsometry(motion,globals) then
                if options.gatherGroup then
                    addIsometry(motion,globals);
                fi;
                
                return motion;
            fi;
        else 
            if motion <> fail then 
                return motion;
            fi;
        fi;
    od;    
    return fail;
end;

FindGramMotion:=function(new,oldList,globals,options)
    local old,motion;
    for old in oldList do
        motion:= gramMotion(new,old,globals,options);
        if  motion<> fail then
            Info(InfoDFS,4,"Found gram motion ",motion, " to ",old.cobasis);
            return true;
        fi;
    od;
    return false;
end;

FindSymmetry:=function(rep,list,globals,options) 
    local stab,ground,leftOut,j,cob1,cob2,starttime,old,groundSize,action;
    starttime:=Runtime();
    
    cob1:=rep.cobasis;
    
    Info(InfoDFS,4,"Looking for orbit for ", rep.cobasis);
    for groundSize in [Size(cob1)+1..globals.rows] do
        
        for old in list do
            
            if ( rep.cobasis=old.cobasis) then 
                profEvent(globals.profile,"DuplicateCobasis");
                profAdd(globals.profile,"DuplicateCobasisTime",
                        Runtime()-starttime);                
                return true;
            fi;
            
            cob2:=old.cobasis;
    
            
            ground:=Union(cob1,cob2);
            leftOut:=Difference([1..globals.rows],ground);
            
            while(Size(ground)< groundSize) do
                j:=PseudoRandom(leftOut);
                Info(InfoDFS,5,"preprocess: adding j = ",j);
                ground:=Union(ground,[j]);
                RemoveSet(leftOut,j);
            od;

            stab:=Stabilizer(globals.PermGrp, ground,OnSets);
            Info(InfoDFS,5,"Stabilizer Size =",Order(stab));
            Info(InfoDFS,5,"Ground =",ground);
            
            action:=RepresentativeAction(stab,cob1,cob2,OnSets);
            if  action <>fail then
                profEvent(globals.profile,"OldCobasis");
                profAdd(globals.profile,"OldCobasisTime",Runtime()-starttime);
                Info(InfoDFS,2,"Found symmetry index = ",old.index,
                     " order = ",Order(stab), 
                     "\tground =", Size(ground), 
                     "\ttime = ",Runtime()-starttime,
                     "\naction =",action,
                     "\nold cobasis=",old.cobasis);
                return true;
            fi;
        od;
    od;
    profEvent(globals.profile,"NewCobasis");
    profAdd(globals.profile,"NewCobasisTime",Runtime()-starttime);
    return false;
end;

# return known cobasis representatives that satisfy the invariants of the 
# given one.

MatchingInvariants:=function(rep,globals,options)
    local matches,old,invariants,preFilter;
    profEvent(globals.profile,"InvariantCheck");
    
    Info(InfoDFS,3,"invariant checking ",rep.cobasis);
    matches:=[];
    
    if (options.hashCobases) then 
        preFilter:=lookupDictionary(globals.cobasisDict,
                           rep.(options.cobHashKey));
        if preFilter=fail then
            return [];
        fi;
    else
        preFilter:=globals.cobasisList;
    fi;
    
    for old in preFilter do 
        invariants:=true;
        if options.cobasisStabilizers then
            invariants:= invariants and (old.stabilizerOrder = rep.stabilizerOrder);
            # Could do something fancier here
#            invariants:= invariants and (old.stabilizerOrbits = rep.stabilizerOrbits);
            
        fi;
        invariants:= invariants and (old.det = rep.det);
        
        if invariants then
            AddSet(matches,old);
        fi;
    od;
    profAdd(globals.profile,"InvariantMatches",Size(matches));
    Info(InfoDFS,4,"Got ",Size(matches), " ,matches");
    return matches;
end;


CanonicalizeCobasis:=function(rep,globals,options)
    
    if options.greedyCanon then
        rep.pseudoCanon:=GreedyRandomCanon(globals.PermGrp,
                                 rep.cobasis,OnSets,
                                 options.greedyTrials);
    elif options.brutalCanon then
        rep.pseudoCanon:=BrutalCanon(rep.cobasis,globals );
    elif options.lazyCanon then
        rep.pseudoCanon:=LazyCanon(rep.cobasis,globals,options );
    else
        rep.pseudoCanon:=fail;
    fi;

    return rep;
end;


OrbitMin:=function(G,element,action)
    local O;
    O:=Orbit(G,element,action);
    return RepresentativeSmallest(O);
end;

CheckByCanons:=function(rep,possibleMatches,globals,options)
    local effort,leftOut,i,j,canon,starttime,newCanons;
    
    starttime:=Runtime();
    profEvent(globals.profile,"CanonCheck");
    
    newCanons:=[];
    leftOut:=Difference([1..globals.rows],rep.cobasis);
    effort:=Minimum(Size(possibleMatches)*options.canonEffort,Size(leftOut));
    for i in [1..effort] do
        j:=PseudoRandom(leftOut);
        RemoveSet(leftOut,j);
        canon:=OrbitMin(globals.stabTable[j],rep.cobasis,OnSets);
        if (canon in globals.pseudoCanons) then
            profAdd(globals.profile,"CanonTime",Runtime()-starttime);
            profEvent(globals.profile,"CanonMatch");
            Info(InfoDFS,2,"Known pseudoCanon ",canon);
            UniteSet(globals.pseudoCanons,newCanons);
            return true;
        fi;
        Info(InfoDFS,2,"Adding pseudoCanon ",canon, " for ", rep.cobasis);        
        AddSet(newCanons,canon);
    od;
    
    Info(InfoDFS,2,"Canon Check Failed");    
    UniteSet(globals.pseudoCanons,newCanons);
    profAdd(globals.profile,"CanonTime",Runtime()-starttime);
    return false;
end;
    

IsNewCobasis:=function(rep,globals,options)
    local old,invariants,newCanon,stab,possibleMatches;
    
    
    
    Info(InfoDFS,2,"checking cobasis ",rep.cobasis,"\n");
    
    Info(InfoDFS,5,"options ",options,"\n");
    

    possibleMatches:=MatchingInvariants(rep,globals,options);
    if Size(possibleMatches)=0 then 
        Info(InfoDFS,2,"New by invariants");
        profEvent(globals.profile,"NewByInvariants");
        return true;
    fi;
                
    
    if (Size(possibleMatches) >=1 and options.trustInvariants) then
       return false;
    fi;
    
    if options.canonTest then
        
        if CheckByCanons(rep,possibleMatches,globals,options) then
            return false;
        fi;
    fi;
    
    if options.gramMotion then
        return not FindGramMotion(rep,possibleMatches,globals,options);
    fi;
    
    if (FindSymmetry(rep,possibleMatches,globals,options)) then 
        return false;
    fi;
    return true;
end;


knownRay:=function(rep,globals,options)
    local old,act;
    Info(InfoDFS,2,"checking ray ",rep.coordinates,"  ");

    for old in globals.rayOrbits do
        act:=RepresentativeAction(globals.PermGrp,rep.incidence,
                    old.incidence,OnSets);
        if  act<>fail then
            Info(InfoDFS,2,"found ", old.coordinates,"\n");
            return old;
        fi;
    od;
    Info(InfoDFS,2,"not found\n");
    return fail;
end;




AddVertex:=function(vertex,globals)
    local normalized;
    Add(globals.vertexOrbits,vertex);
    normalized:=vertex.coordinates/vertex.coordinates[1];
    Info(InfoDFS,1,"adding vertex",normalized);
    AddSet(globals.vertexSet,normalized);
end;

AddCobasis:=function(cob,globals,options)
    local old,new;
    globals.basisCount:=globals.basisCount+1;
    
    cob.index:=globals.basisCount;
    if options.hashCobases then
        if globals.cobasisDict=fail then
            globals.cobasisDict:=newDictionary();
        fi;
        addDictionary(globals.cobasisDict,cob.gramVec,cob);
        
        if (options.debugInvariants) then
            new:=lookupDictionary(globals.cobasisDict,cob.gramVec);
            Info(InfoDFS,3,"Retrieved list size, ",Size(new), " ", IsList(new));
        fi;
    fi;

    if ((not options.gramVec) or (options.debugInvariants)) then
        Add(globals.cobasisList,cob);
    fi;

    
    Info(InfoDFS,3,"adding cobasis ",globals.basisCount, " = ",cob.cobasis,"\n");
    
    if (Size(globals.pseudoCanons) > globals.siftThreshhold) then
        SiftCanons(globals);
    fi;
    
    if cob.pseudoCanon <> fail then
        AddSet(globals.pseudoCanons,cob.pseudoCanon);
        Info(InfoDFS,3,"Adding pseudoCanon ",cob.pseudoCanon);
    fi;

    if (globals.basisCount mod 100 = 0) then 
        Info(InfoDFS,1,globals.basisCount," bases found (", (Runtime()-globals.startTime)," ms)" );
        if (options.gatherGroup) then
            Info(InfoDFS,1,Order(globals.PermGrp)," symmetries found, ",
                 Size(GeneratorsOfGroup(globals.PermGrp)), " generators");
        fi;
    fi;
end;

# Note, it seems to much faster to use the permutation group.
# note, this could probably be sped up by copying the cobasis strategy 
# of looking up in dictionary by gram vector.

knownVertex:=function(rep,globals,options)
    local old,act,normalized,invariants;
    Info(InfoDFS,2,"checking vertex ",rep.coordinates,"  ");
    normalized:=rep.coordinates/rep.coordinates[1];
    if (normalized in globals.vertexSet) then
        Info(InfoDFS,2,"duplicate vertex\n");
        return rep;
    fi;
    
    for old in globals.vertexOrbits do
        invariants:=true;
        if (options.gramVec) then 
            invariants:=invariants and (old.gramVec=rep.gramVec);
        fi;
        
        if (options.debugInvariants or invariants) then
            act:=RepresentativeAction(globals.PermGrp,rep.incidence,
                         old.incidence,OnSets);
            if  act<>fail then
                if (not invariants) then
                    Info(InfoDFS,1,"invariants failed ",
                         rep.incidence,old.incidence,act);
                fi;
                Info(InfoDFS,2,"found ", old.coordinates,"\n");
                return old;
            fi;
        fi;
    od;
    Info(InfoDFS,2,"not found\n");
    return fail;
end;



RayRep:=function(cob,vertexCoords,globals,options)
    local rep;
    
    rep:=rec(incidence:=
             Difference(
                     AsSet(Concatenation(cob.cobasis,cob.extraincidence)),
                     [cob.ray]),
             coordinates:=vertexCoords,det:=cob.det);
    Info(InfoDFS,3,"ray incidence ",rep.incidence);
    if options.gramVec then
        rep.gramVec:=fastGramVec(rep.incidence,globals.innerProductMat);
    fi;
    
    return rep;
end;

GetRays:=function(globals,options)
    local S,C,rep,j;
    for j in [1..globals.real_dim] do
	S:=LRSGetSolution(globals.LRS,j);
 if (S<>fail) then
     Info(InfoDFS,3,"Ray in column ",j);
     C:=LRSGetCobasis(globals.LRS,j);
     Info(InfoDFS,3,"Ray index ",C.ray);
     rep:=RayRep(C,S,globals,options);
     if knownRay(rep,globals,options) =fail then
         Add(globals.rayOrbits,rep);
     fi;
 fi;
od;
end;

#hmm. This might be able to be eliminated. Or at least done faster by LRS.

invertCobasis:=function(cob,M)
    local A,b;
    A:=List(M{cob},p->p{[2..Size(p)]});
    b:=List(M{cob},p->p[1]);
    return Concatenation([1],(1/A) * b);
end;
           
    

CobasisInvariants:=function(cob,vertexCoords,globals,options)
    local rep;
    
    
    rep:=rec(cobasis:=cob.cobasis, 
             extraincidence:=cob.extraincidence,
             coordinates:=vertexCoords,
             det:=cob.det,
             pseudoCanon:=fail,
             stabilizer:=fail
             );
    
    if options.cobasisStabilizers then
        rep.stabilizer:=Stabilizer(globals.PermGrp,cob.cobasis,OnSets);
        rep.stabilizerOrder:=Order(rep.stabilizer);
    fi;
    
    if options.gramVec then
        if options.gramGraph then
            rep.gramGraph:=gramGraph(rep,globals,options);
            rep.gramVec:=gramKey(Size(rep.cobasis),rep.gramGraph);
        else
            rep.gramVec:=fastGramVec(rep.cobasis,globals.innerProductMat);
        fi;
        
    fi;
    
    
    return rep;
end;


    

VertexRep:=function(cob,vertexCoords,globals,options)
    local rep;
    
    rep:=rec(incidence:=SortedList(
                 Concatenation(cob.cobasis,cob.extraincidence)),
             coordinates:=vertexCoords,det:=cob.det);
    
    if options.gramVec then
        rep.gramVec:=fastGramVec(rep.incidence,globals.innerProductMat);
    fi;
    if (options.vertexStabilizers) then
        rep.stabilizer:=Stabilizer(globals.PermGrp,rep.incidences,OnSets);
        rep.stabilizerOrder:=Order(rep.stabilizer);
        rep.stabilizerOrbits:=Orbits(rep.stabilizer,[1..globals.rows],OnPoints);
    else
        rep.stabilizerOrbits:=fail;
    fi;
    
    return rep;
end;


CanonicalizeEntering:=function(entering,rep,globals)
    local orbIndices,keep;
    
    if (entering=[]) then
        return [];
    fi;
    
    orbIndices:=List(entering,j->PositionProperty(rep.stabilizerOrbits,l-> j in l ));
    keep:=Concatenation([1],
                  Filtered([2..Size(entering)],
                          j->orbIndices[j] <> orbIndices[j-1]));
    
    return entering{keep};
end;

SameCoordinates:=function(a,b)
    return a.coordinates=b.coordinates;
end;
PushNewEdges:=function(oldCob,globals,options)
    local entering,enter,leave,newC,C,S,newRep,newVertex,oldVec,cob,sol,
          possibleMatches;
    
    Info(InfoDFS,4,"Checking pivots from ",oldCob);
    for leave in oldCob do
        
        Info(InfoDFS,3,"\tleave = ",leave,"\n");
        if options.lexOnly then
            enter:=LRSLexRatio(globals.LRS,leave);
            if (enter <> fail) then
                entering:=[enter];
            else
                entering:=[];
            fi;
        else
            Info(InfoDFS,4,"\tlrs current cobasis: ", 
                 LRSGetCobasis(globals.LRS,0).orderedcobasis);
            if (options.orbitLexPerturb) then
                entering:=LRSOrbitRatio(globals.LRS,leave);
            else
                entering:=LRSAllRatio(globals.LRS,leave);
                if (Size(entering)>1 and options.slowOrbitRatio) then
                    entering:=slowOrbitRatio(
                                      Position(oldCob,leave),
                                      oldCob,
                                      globals.slackSys,
                                      options.orbitsForPerturbation,
                                      options.lexSigns);
                fi;
            fi;
            
        fi;
#        if (options.cobasisStabilizers)  then
#            entering:=CanonicalizeEntering(entering,rep,globals);
#        fi;        
        
        Info(InfoDFS,3,"\tentering = ",entering,"\n");
        for enter in entering do
            
            Info(InfoDFS,4,"\t\tpivoting ",leave, ",",enter);
            ErrCheck("LRSPivot",LRSPivot(globals.LRS,leave,enter));
            cob:=LRSGetCobasis(globals.LRS,0);
            Info(InfoDFS,4,"\t\tgot ",cob.cobasis);
            sol:=LRSGetVertex(globals.LRS);
            ErrCheck("LRSPivot",LRSPivot(globals.LRS,enter,leave));
            Info(InfoDFS,5,"done\n");
            
            if not lookupCache(globals.cobasisCache,cob) then
                addCache(globals.cobasisCache,cob);
                
                newRep:=CobasisInvariants(cob,sol,globals,options);
                newVertex:=VertexRep(cob,sol,globals,options);
                oldVec:=knownVertex(newVertex,globals,options);
                if oldVec = fail then
                    AddVertex(newVertex,globals);
                    AddCobasis(newRep,globals,options);
                    PushPivot(globals.WorkStack,oldCob,leave,enter);
                elif SameCoordinates(oldVec,newRep) or (not options.dualFacetTrick) then
                    if (IsNewCobasis(newRep,globals,options)) then
                        Info(InfoDFS,2,"pushing ",oldCob," ",leave," ",enter,"\n");
                        PushPivot(globals.WorkStack,oldCob,leave,enter);
                        AddCobasis(newRep,globals,options);
                    fi;
                else
                    # we assume that if the new cobasis is defining a different
                    # vertex, but that vertex is symmetric, then its neighbours
                    # will be symmetric to those of the known vertex.

                    Info(InfoDFS,2,"pruning via the (dual) facet trick\n");
                fi;
            fi;
        od;
    od;
end;



initializeGlobals:=function(options,Grp,Mat)
    local PermGrp,MatGrp,gg,gM,globals,i,lin;
    
    Info(InfoDFS,4,"Initializing Globals");    
    
    globals:=rec(
                 nextIdent:=1,
                 LRS:=fail,
                 Nauty:=fail,
                 dim:=Size(Mat[1]),
                 rows:=Size(Mat),
                 basisCount:=0,
                 siftThreshhold:=options.siftThreshhold,
                 startTime:=Runtime(),
                 cobasisList:=[],
                 WorkStack:=[],
                 PathStack:=[],
                 rayOrbits:=[],
                 vertexOrbits:=[],
                 vertexSet:=[],
                 pseudoCanons:=[],
                 slackSys:=rec(),
                 Mat:=Mat,
                 cobasisCache:=newCache(options.cacheSize),
                 profile:=newProfile(),
                 isometryCache:=newCache(options.cacheSize),
                 stabTable:=[]);

    if (options.gramVec) then
        if (options.innerProductMat <> fail) then
            globals.innerProductMat:=options.innerProductMat;
        else
            globals.innerProductMat:=InnerProdMat([1..Size(Mat)],Mat,
                                             options.InnerProduct);
        fi;
        globals.cobasisDict:=fail;
    fi;
    
    # this should eventually go away when working in lrsserv
    if (options.slowOrbitRatio) then
        globals.slackSys:=slackSys(globals.Mat);
    fi;
    globals.LRS:=LRS();
    globals.Nauty:=Nauty();
    
    Info(InfoDFS,4,"Initializing Double Group Representation...");
    if IsMatrixGroup(Grp) then
        MatGrp:=Grp;
        PermGrp:=PermGroupMatGroup(Grp,Mat);
    elif IsPermGroup(Grp) then
        PermGrp:=Grp;
#        MatGrp:=MatGroupPermGroup(Grp,Mat);
    elif (options.nautyIsometries) then
        Info(InfoDFS,1,"No Group Provided. Computing isometries");        
        PermGrp:=groupViaGram3(Mat,DefaultInnerProduct,globals.Nauty);
 #       MatGrp:=MatGroupPermGroup(PermGrp,Mat);
        Info(InfoDFS,1,"Found Group of order ",Order(PermGrp));    
        globals.nauty_preproc_time:=NautyRUsage(globals.Nauty);
    else
            
        PermGrp:=Group(());
#        MatGrp:=Group(IdentityMat(Size(Mat[1])));
        options.gatherGroup:=true;
        Info(InfoDFS,1,"No Group Provided. Switching to gather mode.");
    fi;
        
    Info(InfoDFS,4,"Done.");
    globals.PermGrp:=PermGrp;
#    globals.MatGrp:=MatGrp;
        
    if (IsGroup(PermGrp)) then
        Info(InfoDFS,4,"Computing Input Orbits");        
        globals.inputOrbits:=Orbits(PermGrp,[1..globals.rows],OnPoints);
#        for i in [1..globals.rows] do
 #           globals.stabTable[i]:=Stabilizer(PermGrp,
  #                                        Difference([1..globals.rows],[i]),OnSets);
  #      od;
    fi;
    if (options.lazyCanon) then
        globals.groundMax:=globals.dim+1;
    else
        globals.groundMax:=globals.rows;
    fi;
    
    
    if (options.linearities = false ) then
        lin:=[];
    else
        lin:=options.linearities;
    fi;
    
    globals.linearities:=lin;

    return globals;
end;

initializeOptions:=function(rawOptions)
    local options,opt;
    
    Info(InfoDFS,4,"Initializing Options");
    options:=rec(
                 InnerProduct:=DefaultInnerProduct,
                 VRepresentation:=false,
                 basisLimit:=infinity,
                 brutalCanon:=false,
                 cacheSize:=1000,
                 canonEffort:=1,
                 canonTest:=false,
                 cobHashKey:=fail,
                 cobasisStabilizers:=false,
                 debugInvariants:=false,
                 dualFacetTrick:=true,
                 gatherGroup:=false,
                 gramGraph:=false,
                 gramMotion:=false,
                 gramMotionViaBrute:=false,
                 gramMotionViaGroup:=true,
                 gramVec:=true,
                 gramViaStabilizer:=false,
                 greedyCanon:=false,
                 greedyTrials:=200,
                 hashCobases:=false,
                 infoLevel:=1,
                 innerProductMat:=fail,
                 lazyCanon:=false,
                 lexOnly:=false,
                 lexSigns:=fail,
                 linearities:=[],
                 nautyIsometries:=false,
                 orbitLexPerturb:=false,
                 orbitsForPerturbation:=fail,
                 restartFrom:="first",
                 siftThreshhold:=1000,
                 slowOrbitRatio:=false,
                 startingCobasis:=fail,
                 trustCanon:=true,
                 trustInvariants:=false,
                 vertexStabilizers:=false,
                 );
                 
    for opt in RecNames(rawOptions) do
        options.(opt):=rawOptions.(opt);
    od;
    
    if (options.gramVec) then
        options.hashCobases:=true;
        options.cobHashKey:="gramVec";
    fi;
    
    if (options.gramMotionViaBrute) then
        options.gramMotionViaGroup:=false;
    fi;
    return options;
end;
    

# caller is responsible for pivoting to the right dictionary before 
# calling
dfsFromRoot:=function(root,globals,options)
    local cob,cobList,l,pivot,btpivot,dict,rval;
    
    l:=globals.LRS;
    
    PushNewEdges(root,globals,options);    
    
    while (Size(globals.WorkStack)>0  
           and options.basisLimit > globals.basisCount) do
        dict:=LRSGetCobasis(l,0);
        if (options.breadthFirst) then
            pivot:=DequeuePivot(globals.WorkStack);
            if (dict.cobasis <> pivot.cobasis) then
                # this is a hack, there should be a better backtracking
                LRSSetCobasis(l,pivot.cobasis);
                dict:=LRSGetCobasis(l,0);
            fi;
        else
            pivot:=PopPivot(globals.WorkStack);
            while (dict.cobasis <> pivot.cobasis) do
                btpivot:=PopPivot(globals.PathStack);
                ErrCheck("LRSPivot",LRSPivot(l,btpivot.leave,btpivot.enter));
                dict:=LRSGetCobasis(l,0);
            od;
            
        fi;
            
    	
        Info(InfoDFS,2,"pivot ",dict.cobasis,",",pivot.leave,",",pivot.enter,"\n");
        ErrCheck("LRSPivot",LRSPivot(l,pivot.leave,pivot.enter));
        
        cob:=LRSGetCobasis(l,0);
        GetRays(globals,options);
        PushNewEdges(cob.cobasis,globals,options);
        
        if (not options.breadthFirst) then
            PushPivot(globals.PathStack,[],pivot.enter,pivot.leave);
        fi;
    od;
    if options.gramVec then
        cobList:=Concatenation(globals.cobasisDict.data);
    else
        cobList:=globals.cobasisList;
    fi;
    rval:=rec(dimension:=Size(globals.Mat[1])-1,
              vertex_orbits:=globals.vertexOrbits,
               basis_orbits:=cobList,
               pseudoCanons:=globals.pseudoCanons,
               profile:=globals.profile,
              ray_orbits:=globals.rayOrbits,
              symmetry_generators:=GeneratorsOfGroup(globals.PermGrp));
    if (IsBound(globals.nauty_preproc_time)) then
        rval.nauty_preproc_time:=globals.nauty_preproc_time;
    fi;
    if (options.basisLimit > globals.basisCount) then
        rval.finished:=true;
    else
        Info(InfoDFS,1,"Basis Limit of ",options.basisLimit," hit. Terminating");
        rval.finished:=false;
    fi;

    return rval;
end;

dfsFirstBasis:=function(globals,options)
    local cob,sol,rep,l,result;
    l:=globals.LRS;
    
    result:=LRSGetFirstBasis(l);
    if (result=fail) then
        Error("LRSGetFirstBasis");
    else
        globals.real_dim:=result.dim;
    fi;
    
    globals.lrs_linearities:=LRSGetLinearities(l);
    Info(InfoDFS,1,"Found  ",Size(globals.lrs_linearities)," Linearities");

    if (options.startingCobasis <> fail) then
        LRSSetCobasis(l,options.startingCobasis);
    fi;
    
    cob:=LRSGetCobasis(l,0);
    sol:=LRSGetVertex(l);
    rep:=CobasisInvariants(cob,sol,globals,options);
    rep:=CanonicalizeCobasis(rep,globals,options);
    Info(InfoDFS,2,"initial cobasis = ",rep);
    globals.initialCobasis:=rep;
    AddCobasis(rep,globals,options);
    AddVertex(VertexRep(cob,sol,globals,options),globals);
    GetRays(globals,options);
    return cob;
end;

CopyGlobals:=function(globals,options,pass2options)
    local cobList,cob,rep,pass2globals;
    
    # probably some cleaning up is needed here
    pass2globals:=globals;
    
    pass2globals.PathStack:=[];
    pass2globals.pseudoCanons:=[];
    pass2globals.cobasisCache:=newCache(pass2options.cacheSize);
    
    if (options.hashCobases) then
        cobList:=Concatenation(globals.cobasisDict.data);
    else
        cobList:=globals.cobasisList;
    fi;
    
    if pass2options.restartFrom="all" then
        pass2globals.cobasisQueue:=List(cobList, r->r.cobasis);
    else
        if pass2options.restartFrom="first" then
            cob:=cobList[1].cobasis;
        else
            cob:=Random(cobList).cobasis;
        fi;
        pass2globals.cobasisQueue:=[cob];
    fi;
    
    if (pass2options.hashCobases) then

        if (pass2options.cobHashKey=options.cobHashKey) then
            pass2globals.cobasisDict:=globals.cobasisDict;
        else
            for rep in cobList do
                AddCobasis(rep,pass2globals,pass2options);
            od;
        fi;
        
    else
        pass2globals.cobasisList:=cobList;
    fi;
    
    return pass2globals;
end;

dfs:=function(Mat,Grp,optionsList)
    local l,  dict,  globals,
          options,opt,cob,results,pass,oldOptions,oldGlobals;
    
    
    
    for pass in [1..Size(optionsList)] do
        Info(InfoDFS,1,"Starting pass ", pass);
        
        if (pass = 1) then 
            options:=initializeOptions(optionsList[1]);
            SetInfoLevel(InfoDFS,options.infoLevel);            
            
            globals:=initializeGlobals(options,Grp,Mat);    
            
            l:=globals.LRS;
            LRSSetOptions(l,options);
            LRSSetGeometric(l);
            LRSLoadMatrix(l,Mat,globals.linearities);
            if (options.orbitsForPerturbation <> fail) then
                LRSOrbits(l,  options.orbitsForPerturbation);
            else
                LRSOrbits(l,  globals.inputOrbits);
            fi;
            if (options.lexSigns <> fail) then
                LRSLexSigns(l,options.lexSigns);
            fi;

            cob:=dfsFirstBasis(globals,options);
            

            globals.cobasisQueue:=[cob.cobasis];
        else
            oldOptions:=options;
            oldGlobals:=globals;
                
            options:=initializeOptions(optionsList[pass]);
            globals:=CopyGlobals(oldGlobals,oldOptions,options);
            Unbind(oldGlobals);
            Unbind(oldOptions);
            
        fi;
    
        for cob in globals.cobasisQueue do
            Info(InfoDFS,1,"starting from ",cob);            
            LRSSetCobasis(l,cob);
            results:=dfsFromRoot(cob,globals,options);
        od;
        
    od;
    results.initial_cobasis:=globals.initialCobasis.cobasis;
    results.nauty_total_time:=NautyRUsage(globals.Nauty);
    CloseStream(globals.LRS);
    CloseStream(globals.Nauty);
    results.gap_time:=Runtimes();

    return results;
end;            

DumpRec:=function(out, print_keys)
    local key;
    for key in print_keys do
        if IsBound(out.(key)) then
            if (IsRecord(out.(key))) then
                Print(key," => { \n");
                DumpRec(out.(key),RecNames(out.(key)));
                Print("},\n");
            else
                Print(key," => ", out.(key),",\n");
            fi;
        fi;
    od;
                
end;

CountRec:=function(out, print_keys)
    local key;
    for key in print_keys do
        if IsBound(out.(key)) then
            if (IsList(out.(key))) then
                Print(key," => ",Size(out.(key)),",\n");
            fi;
        fi;
    od;
                
end;


FormatResults:=function(out,key)
    Print("\n#STARTRESULTS\n");
    Print("{\n");
    Print("key => '",key,"',\n");
    DumpRec(out,["dimension","gap_time","nauty_preproc_time",
            "nauty_total_time","initial_cobasis"]
            );
    
    Print("group_order =>",Order(Group(out.symmetry_generators)),",\n");;
    CountRec(out,["basis_orbits","vertex_orbits",
            "ray_orbits","symmetry_generators"]);
    Print("}\n");
    Print("\n#ENDRESULTS\n");

end;
